/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet remote device and sensor
 *
 * Chip type           : ATMEGA168/atmega328 with ENC28J60
 *********************************************/
#define F_CPU 12500000UL  // 12.5 MHz
#include <util/delay.h>
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "avr_compat.h"
#include "net.h"
#include "sensirion_protocol.h"

#define VERSIONSTRING "<br><hr>version 3.3\n"
// number of sensors (exmple: if you connect PD0 and PD1 then set to 2)
#define MAX_SENSORS 2

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
//static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x01,0x27};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
//static uint8_t myip[4] = {192,168,1,27};

static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x26};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,26};
//
// listen port for tcp/www 
#define MYWWWPORT 80 
// or on a different port:
// listen port for tcp/www 
//#define MYWWWPORT 88  
//
// bigger pages need more buffer but we are limited be the amount of ram
#define BUFFER_SIZE 560
static uint8_t buf[BUFFER_SIZE+1];

// uncomment to see raw sensor data:
//#define SHOW_RAW 1

unsigned int gHumival_raw[MAX_SENSORS],gTempval_raw[MAX_SENSORS];
uint8_t gSensorErrors[MAX_SENSORS];
uint8_t gAllSensorsReadOnce=0;
//
static volatile uint8_t cnt2step=0;
static volatile uint8_t gSec=0;

// gerneral purpose LED
/* set output to Vcc, red LED off */
#define LED1OFF PORTB|=(1<<PORTB1)
/* set output to GND, red LED on */
#define LED1ON PORTB&=~(1<<PORTB1)
// 
// 

// convert temperature to kelvin times 10
int c2f(int celsiustimes10){
        return((celsiustimes10 * 18 + 3200)/10);
}

// insert a '.' before the last digit/character in the string.
// The user is responsible for allocating enough space.
// "100" will become "10.0"
// Special care must be taken if there is just one digit: -5-> -0.5
void adddecimalpoint(char *str){
        char c;
        if (*str == '-'){
                str++;
        }
        if (*(str+1) == '\0'){
                // just one digit
                c=*str;
                *str='0';
                *(str+1)='.';
                *(str+2)=c;
                *(str+3)='\0';
                return;
        }
        while(*str){
                c=*str;
                str++;
                if (!*str){
                        // last char, insert a dot
                        *(str-1)='.';
                        *str=c;
                        str++;
                        *str='\0';
                }
        }
}

// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        uint8_t error;
        uint8_t rh,i;
        int dew,temp; 
        char str[15];
        char snum[5];
        // autmatically reload (refresh) page every 5 min:
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\nRefresh: 300\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Sensors @"));
        itoa(myip[3],str,10);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>\n"));
        i=0;
        if (gAllSensorsReadOnce==0){
                plen=fill_tcp_data_p(buf,plen,PSTR("wait...\n"));
                goto END_OF_WEBPAGE;
        }
        while(i<MAX_SENSORS){
                //
                snum[0]='s';
                itoa(i,&snum[1],10);
                snum[2]=' ';
                snum[3]='\0';
                error=gSensorErrors[i];
                if(error!=0){ //in case of an error
                        plen=fill_tcp_data(buf,plen,snum);
                        if (error==1){
                                plen=fill_tcp_data_p(buf,plen,PSTR("no sensor\n"));
                        }
                        if (error==2){
                                plen=fill_tcp_data_p(buf,plen,PSTR("sensor crc error\n"));
                        }
                        if (error==3){
                                plen=fill_tcp_data_p(buf,plen,PSTR("timeout error\n"));
                        }
                        plen=fill_tcp_data_p(buf,plen,PSTR("\n"));
                        i++;
                        continue;
                }
#ifdef SHOW_RAW
                plen=fill_tcp_data(buf,plen,snum);
                plen=fill_tcp_data_p(buf,plen,PSTR("DEBUG: raw temp humi:"));
                itoa(gTempval_raw[i],str,10);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR(" "));
                itoa(gHumival_raw[i],str,10);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("\n"));
#endif
                //
                plen=fill_tcp_data(buf,plen,snum);
                plen=fill_tcp_data_p(buf,plen,PSTR("temperature: "));
                temp=calc_sth11_temp(gTempval_raw[i]);
                itoa(temp,str,10);
                adddecimalpoint(str);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("'C ["));
                itoa(c2f(temp),str,10);
                adddecimalpoint(str);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("'F]"));
                plen=fill_tcp_data_p(buf,plen,PSTR("\n"));
#ifdef SHOW_RAW
                plen=fill_tcp_data(buf,plen,snum);
                plen=fill_tcp_data_p(buf,plen,PSTR("humi lin   : "));
                rh=rhcalc_int(gHumival_raw[i]);
                itoa(rh,str,10);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("\n"));
#endif
                plen=fill_tcp_data(buf,plen,snum);
                plen=fill_tcp_data_p(buf,plen,PSTR("humidity   : "));
                rh=calc_sth11_humi(gHumival_raw[i],temp);
                itoa(rh,str,10);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("%\n"));
                plen=fill_tcp_data(buf,plen,snum);
                plen=fill_tcp_data_p(buf,plen,PSTR("dew point  : "));
                dew=calc_dewpoint(rh,temp);
                itoa(dew,str,10);
                adddecimalpoint(str);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("'C ["));
                // cut comma as the comma digits are not accurate:
                itoa(c2f(dew)/10,str,10);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("'F]\n\n"));
                //
                i++;
        }
END_OF_WEBPAGE:
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=\"./\">refresh page</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR(VERSIONSTRING));
        return(plen);
}

// called when TCNT2==OCR2A
// that is in 50Hz intervals
ISR(TIMER2_COMPA_vect){
        cnt2step++;
        if (cnt2step>49){ 
                gSec++;
                cnt2step=0;
        }
}

/* setup timer T2 as an interrupt generating time base.
* You must call once sei() in the main program */
void init_cnt2(void)
{
        cnt2step=0;
        PRR&=~(1<<PRTIM2); // write power reduction register to zero
        TIMSK2=(1<<OCIE2A); // compare match on OCR2A
        TCNT2=0;  // init counter
        OCR2A=244; // value to compare against
        TCCR2A=(1<<WGM21); // do not change any output pin, clear at compare match
        // divide clock by 1024: 12.5MHz/1024=12207.0313 Hz
        TCCR2B=(1<<CS22)|(1<<CS21)|(1<<CS20); // clock divider, start counter
        // 12207.0313 / 244= 50.0288
}

int main(void){
        uint16_t plen;
        uint16_t dat_p;
        uint8_t i=0;
        uint8_t measurementstate=0;
        uint8_t sensor=0;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_ms(1);

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_ms(10);
        
        // enable PB1, LED as output 
        DDRB|= (1<<DDB1);
        LED1OFF;

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PORTD7);// transistor off
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_ms(20);
        
        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,MYWWWPORT);

        // time keeping
        init_cnt2();
        sei();

        i=0;
        while(i<MAX_SENSORS){
                gSensorErrors[i]=1; // init as error
                i++;
        }
        while(1){
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will ne unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        // no packets, do measurements
                        //
                        // we need the measurementstate otherwise we would
                        // execute the same thing serveral times in that second
                        if (measurementstate==0 && gSec==0){
                                s_connectionreset(sensor);
                                LED1ON;
                                measurementstate++;
                        }
                        if (measurementstate==1 && gSec==1){
                                gSensorErrors[sensor]=s_measure(MEASURE_TEMP,sensor); //measure temperature
                                measurementstate++;
                                LED1OFF;
                                if (gSensorErrors[sensor]!=0){
                                        measurementstate=5;
                                }
                        }
                        if (measurementstate==2) {
                                // now poll sensor until it is ready
                                // might take up to 220ms
                                if (s_resultready(sensor)){
                                        gSensorErrors[sensor]=s_readmeasurement( &gTempval_raw[sensor],sensor); 
                                        measurementstate++;
                                        if (gSensorErrors[sensor]!=0){
                                                measurementstate=5;
                                        }
                                }
                        }
                        if (gSec==2){
                                if (measurementstate==2){
                                        gSensorErrors[sensor]=3; // timeout
                                        measurementstate=5;
                                }
                                if (measurementstate==3){

                                        gSensorErrors[sensor]=s_measure(MEASURE_HUMI,sensor); 
                                        measurementstate++;
                                        if (gSensorErrors[sensor]!=0){
                                                measurementstate=5;
                                        }
                                }
                        }
                        if (measurementstate==4) {
                                // now poll sensor until it is ready
                                // might take up to 220ms
                                if (s_resultready(sensor)){
                                        gSensorErrors[sensor]=s_readmeasurement( &gHumival_raw[sensor],sensor);
                                        measurementstate++;
                                }
                        }
                        // wrap at 3:
                        if (gSec>2){
                                gSec=0;
                                measurementstate=0;
                                sensor++;
                                // start over
                                if (sensor>=MAX_SENSORS){
                                        sensor=0;
                                        gAllSensorsReadOnce=1;
                                }
                        }
                        //
                        continue;
                }
                        
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to 
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        make_arp_answer_from_request(buf);
                        continue;
                }

                // check if ip packets are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
                // tcp port www start, compare only the lower byte
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V&&buf[TCP_DST_PORT_H_P]==0&&buf[TCP_DST_PORT_L_P]==MYWWWPORT){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                                        // head, post and other methods:
                                        //
                                        // for possible status codes see:
                                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                                        goto SENDTCP;
                                }
                                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                                        plen=print_webpage(buf);
                                        goto SENDTCP;
                                }
                                // not the correct url. In this case we 
                                // just accept the base url (hostname/ip only).
                                // Generate and error.
                                //
                                // for possible status codes see:
                                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                // 401 Unauthorized
                                // 403 Forbidden
                                // 404 Not Found
                                // 410 Gone
                                //
                                //plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
                                plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                                //
SENDTCP:
                                make_tcp_ack_from_any(buf); // send ack for http get
                                make_tcp_ack_with_data(buf,plen); // send data
                                continue;
                        }

                }
                // tcp port www end
                //
        }
        return (0);
}
